/* Distributed Checksum Clearinghouse
 *
 * Copyright (c) 2005 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC and permission to use
 * U.S. Patent 6,330,590 by contacting Commtouch at http://www.commtouch.com/
 * or by email to nospam@commtouch.com.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.16-1.42 $Revision$
 */

#include "dcc_ck.h"


static char *
ck_word(char *p,
	const char *word,
	int word_len)
{
	int sps;

	if (strncasecmp(p, word, word_len))
		return 0;

	p += word_len;
	sps = strspn(p, DCC_WHITESPACE);
	if (sps == 0)
		return 0;

	return p+sps;
}



int					/* -1=fatal 0=problems 1=ok */
dcc_parse_whitefile(DCC_EMSG emsg,
		    DCC_WF *wf,
		    const DCC_PATH white_nm_base,
		    FILE *white_file,	/* stdio for white_nm_base */
		    DCC_PARSED_CK_FNC add_fnc,
		    DCC_PARSED_CIDR_CK_FNC cidr_fnc)
{
	char buf[200];
	char *type, *ck, *p;
	DCC_TGTS new_tgts;
	const char *white_nm;
	u_int white_nm_num = 0;
	DCC_PATH cur_white_nm;
	FILE *cur_white_file, *new_white_file;
	int main_lineno, lineno;
	u_char result, hex;
	int i;

	result = 1;
	cur_white_file = white_file;
	white_nm = white_nm_base;
	main_lineno = lineno = 0;
	new_tgts = DCC_TGTS_INVALID;
	for (;;) {
		/* Each substantive line has one of the forms:
		 *	tgts	[hex] type	string
		 *		[hex] type	string
		 *	include	pathname
		 *	option ...
		 * A missing number of targets means the line has the
		 * same number of targets as the previous line */

		if (!fgets(buf, sizeof(buf), cur_white_file)) {
			if (ferror(cur_white_file))
				dcc_pemsg(EX_IOERR, emsg,
					  "fgets(whitelist %s): %s",
					  DCC_NM2PATH(white_nm), ERROR_STR());

			/* at the end of an included file,
			 * return to the main file */
			if (cur_white_file != white_file) {
				fclose(cur_white_file);
				cur_white_file = white_file;
				white_nm = white_nm_base;
				lineno = main_lineno;
				continue;
			}
			return result;
		}

		++lineno;

		/* Ignore blank lines and lines starting with '#' */
		type = buf+strspn(buf, DCC_WHITESPACE);
		if (*type == '\0' || *type == '#')
			continue;

		/* trim trailing blanks */
		p = type+strlen(type);
		while (--p > type) {
			if (*p != ' ' && *p != '\t'
			    && *p != '\n' && *p != '\r')
				break;
			*p = '\0';
		}

		/* parse
		 *	include	pathname */
		p = ck_word(type, "include", STRZ("include"));
		if (p) {
			if (cur_white_file != white_file) {
				dcc_pemsg(EX_DATAERR, emsg,
					  "nested \"include\" not allowed%s",
					  fnm_lineno(white_nm_base, lineno));
				result = 0;
				continue;
			}

			/* trim quotes if present from the file name */
			i = strlen(p);
			if (i > 1
			    && ((p[0] == '"' && p[i-1] == '"')
				|| (p[0] == '<' && i > 1 && p[i-1] == '>'))) {
				p[i-1] = '\0';
				++p;
				i -= 2;
			}

			if (i == 0 || i >= ISZ(cur_white_nm)) {
				dcc_pemsg(EX_DATAERR, emsg,
					  "unrecognized \"include %s\"%s",
					  p, fnm_lineno(white_nm, lineno));
				result = 0;
				continue;
			}
			if (white_nm_num >= DIM(wf->info->hdr.white_incs)) {
				dcc_pemsg(EX_DATAERR, emsg,
					  "too many \"include\" files%s",
					  fnm_lineno(white_nm_base, lineno));
				result = 0;
				continue;
			}
			new_white_file = fopen(p, "r");
			if (!new_white_file) {
				dcc_pemsg(EX_DATAERR, emsg,
					  "\"include %s\": %s%s",
					  p, ERROR_STR(),
					  fnm_lineno(white_nm_base, lineno));
				result = 0;
				continue;
			}

			if (wf->info) {
				struct stat sb;

				BUFCPY(wf->info->hdr.white_incs[white_nm_num
							].nm,
				       p);
				if (0 > fstat(fileno(new_white_file), &sb)) {
					wf->info->hdr.white_incs[white_nm_num
							].mtime = 0;
				} else {
					wf->info->hdr.white_incs[white_nm_num
							].mtime = sb.st_mtime;
				}
			}
			++white_nm_num;
			strcpy(cur_white_nm, p);
			white_nm = cur_white_nm;
			cur_white_file = new_white_file;
			main_lineno = lineno;
			lineno = 0;
			continue;
		}

		/* honor greylist controls in client whitelists of the forms:
		 *	option log-{all,normal}
		 *	option greylist-{on,off,log-on,log-off}
		 *	option DCC-{on,off}
		 *	option DCC-reps-{on,off}
		 *	option dnsbl-{on,off}
		 *	option xfltr-{on,off}
		 */
		p = ck_word(type, "option", STRZ("option"));
		if (p) {
			static const struct {
			    const char *str; int len; u_int on; u_int off;
			} *tp, tbl[] = {
#			define DE(s,on,off) {s,STRZ(s),on,~(off)},
				DE("log-all",
				   DCC_WHITE_FG_LOG_ALL,
				   0)
				DE("log-normal",
				   DCC_WHITE_FG_LOG_NORMAL,
				   0)

				DE("greylist-on",
				   DCC_WHITE_FG_GREY_ON,
				   DCC_WHITE_FG_GREY_OFF)
				DE("greylist-off",
				   DCC_WHITE_FG_GREY_OFF,
				   DCC_WHITE_FG_GREY_ON)

				DE("greylist-log-on",
				   DCC_WHITE_FG_GREY_LOG_ON,
				   DCC_WHITE_FG_GREY_LOG_OFF)
				DE("greylist-log-off",
				   DCC_WHITE_FG_GREY_LOG_OFF,
				   DCC_WHITE_FG_GREY_LOG_ON)

				DE("DCC-on",
				   DCC_WHITE_FG_DCC_ON,
				   DCC_WHITE_FG_DCC_OFF)
				DE("DCC-off",
				   DCC_WHITE_FG_DCC_OFF,
				   DCC_WHITE_FG_DCC_ON)

				DE("MTA-first",
				   DCC_WHITE_FG_MTA_FIRST,
				   DCC_WHITE_FG_MTA_LAST)
				DE("MTA-last",
				   DCC_WHITE_FG_MTA_LAST,
				   DCC_WHITE_FG_MTA_FIRST)

				DE("DCC-reps-on",
				   DCC_WHITE_FG_DCC_REP_ON,
				   DCC_WHITE_FG_DCC_REP_OFF)
				DE("DCC-reps-off",
				   DCC_WHITE_FG_DCC_REP_OFF,
				   DCC_WHITE_FG_DCC_REP_ON)

				DE("DNSBL-on",
				   DCC_WHITE_FG_DNSBL_ON,
				   DCC_WHITE_FG_DNSBL_OFF)
				DE("DNSBL-off",
				   DCC_WHITE_FG_DNSBL_OFF,
				   DCC_WHITE_FG_DNSBL_ON)

				DE("xfltr-on",
				   DCC_WHITE_FG_XFLTR_ON,
				   DCC_WHITE_FG_XFLTR_OFF)
				DE("xfltr-on",
				   DCC_WHITE_FG_XFLTR_OFF,
				   DCC_WHITE_FG_XFLTR_ON)
#			undef DE
			};

			if (!wf->info) {
				dcc_pemsg(EX_DATAERR, emsg, "\"option\""
					  " not legal in server whitelist%s",
					  fnm_lineno(white_nm, lineno));
				result = 0;
				continue;
			}
			i = strlen(p);
			for (tp = tbl; ; ++tp) {
				if (tp > LAST(tbl)) {
					dcc_pemsg(EX_DATAERR, emsg,
						  "unrecognized option%s",
						  fnm_lineno(white_nm, lineno));
					result = 0;
					break;
				}
				if (i == tp->len && !strcasecmp(tp->str, p)) {
					wf->info_flags |= tp->on;
					wf->info->hdr.flags |= tp->on;
					wf->info_flags &= tp->off;
					wf->info->hdr.flags &= tp->off;
					break;
				}
			}
			continue;
		}

		/* honor logging controls in the obsolete form
		 *	log {all-grey,no-grey}
		 */
		p = ck_word(type, "log", STRZ("log"));
		if (p) {
			if (!wf->info) {
				dcc_pemsg(EX_DATAERR, emsg, "\"log\""
					  " not legal in server whitelist%s",
					  fnm_lineno(white_nm, lineno));
				result = 0;
			} else if (!strcasecmp(p, "all-grey")) {
				wf->info_flags |= DCC_WHITE_FG_GREY_LOG_ON;
				wf->info->hdr.flags |= DCC_WHITE_FG_GREY_LOG_ON;
			} else if (!strcasecmp(p, "no-grey")) {
				wf->info_flags |= DCC_WHITE_FG_GREY_LOG_OFF;
				wf->info->hdr.flags |=DCC_WHITE_FG_GREY_LOG_OFF;
			} else {
				dcc_pemsg(EX_DATAERR, emsg, "syntax error%s",
					  fnm_lineno(white_nm, lineno));
				result = 0;
			}
			continue;
		}


		/* Look for the number of targets in a simple line */
		if (type != buf) {
			/* If the line started with white space, the number
			 * of targets is the same as the previous line. */
			buf[0] = '\0';
		} else {
			type += strcspn(type, DCC_WHITESPACE);
			if (*type == '\0') {
				dcc_pemsg(EX_DATAERR, emsg,
					  "missing type in \"%s\"%s",
					  buf, fnm_lineno(white_nm, lineno));
				result = 0;
				continue;
			}
			*type++ = '\0';
			/* buf now starts with null-terminated
			 * number of targets, "include", or "log" */
			type += strspn(type, DCC_WHITESPACE);
		}

		ck = type+strcspn(type, DCC_WHITESPACE);

		if (*ck != '\0') {
			/* null terminate the type */
			*ck++ = '\0';
			ck += strspn(ck, DCC_WHITESPACE);
		}

		if (strcasecmp(type, "hex")) {
			hex = 0;
		} else {
			hex = 1;
			type = ck;
			ck = type+strcspn(type, DCC_WHITESPACE);
			if (*ck != '\0') {
				*ck++ = '\0';
				ck += strspn(ck, DCC_WHITESPACE);
			}
		}

		/* parse the target count if it is present instead of blank */
		if (buf[0] != '\0')
			new_tgts = dcc_str2cnt(buf);
		if (new_tgts == 0 || new_tgts == DCC_TGTS_INVALID) {
			dcc_pemsg(EX_DATAERR, emsg,
				  "missing or invalid # of targets \"%s\"%s",
				  buf, fnm_lineno(white_nm, lineno));
			new_tgts = DCC_TGTS_INVALID;
			result = 0;
			continue;
		}

		if (*ck == '\0') {
			dcc_pemsg(EX_DATAERR, emsg, "missing value%s",
				  fnm_lineno(white_nm, lineno));
			result = 0;
			continue;
		}

		/* Look for the type of the checksum, compute the checksum,
		 * and write the checksum.  If it is a host name, write
		 * all of its aliases to the hash table. */
		if (hex) {
			i = dcc_parse_hex_ck(emsg, wf, white_nm, lineno,
					     type, ck, new_tgts,
					     add_fnc);
		} else {
			i = dcc_parse_ck(emsg, wf, white_nm, lineno,
					 type, ck, new_tgts,
					 add_fnc, cidr_fnc);
		}
		/* give up now on a fatal problem adding a
		 * checksum to the file */
		if (i < 0)
			break;
		if (i == 0)
			result = 0;
	}

	/* failed */
	if (cur_white_file && cur_white_file != white_file)
		fclose(cur_white_file);
	return -1;
}
