/* Distributed Checksum Clearinghouse
 *
 * threaded version of client library
 *
 * Copyright (c) 2005 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC and permission to use
 * U.S. Patent 6,330,590 by contacting Commtouch at http://www.commtouch.com/
 * or by email to nospam@commtouch.com.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.16-1.68 $Revision$
 */


#include "cmn_defs.h"
#include "dcc_paths.h"

const char *tmpdir;

CMN_ACTION action = CMN_REJECT;

CHGHDR chghdr = SETHDR;

const char *userdirs;
static int userdirs_len;

u_char dcc_query_only = 0;
u_char try_extra_hard = 0;		/* 1=don't quit if DCC server dead */
u_char to_white_only = 0;
const char *mapfile_nm = DCC_MAP_NM_DEF;

u_int dcc_ctxt_sn = 1;			/* change X-DCC header server name */

RCPT_ST *rcpt_st_free;

static pthread_mutex_t work_mutex;

/* lock al CWF structures as well as the cmn_wf structure */
static pthread_mutex_t cwf_mutex;

typedef struct cwf {			/* private whitelist state */
    struct cwf *older, *newer;
    DCC_WF	wf;
} CWF;
static CWF *cur_cwf, cwfs[NUM_CWFS];


void
cmn_init(DCC_EMSG emsg)
{
	/* Some pthreads implementations (e.g. AIX) don't like static
	 * POSIX thread initializations */
	pthread_mutex_init(&work_mutex, 0);

	/* start the client library threads and locks */
	dcc_clnt_thread_init();
	for (cur_cwf = cwfs; cur_cwf <= LAST(cwfs); ++cur_cwf) {
		cur_cwf->newer = cur_cwf-1;
		cur_cwf->older = cur_cwf+1;
		dcc_wf_init(&cur_cwf->wf, DCC_WF_PER_USER);
	}
	cur_cwf = cwfs;
	LAST(cwfs)->older = cur_cwf;
	cur_cwf->newer = LAST(cwfs);
	dcc_mutex_init(&cwf_mutex, "cwf_mutex");

	if (!dcc_map_info(emsg, mapfile_nm, -1))
		dcc_logbad(EX_USAGE, "%s", emsg);

	dcc_wf_init(&cmn_wf, 0);
	if (main_white_nm
	    && !dcc_new_white_nm(emsg, &cmn_wf, main_white_nm)) {
		dcc_error_msg("%s", emsg);
		main_white_nm = 0;
	}
	if (main_white_nm) {
		switch (dcc_rdy_white(emsg, &cmn_wf)) {
		case DCC_WHITE_OK:
		case DCC_WHITE_NOFILE:
		case DCC_WHITE_SILENT:
			break;
		case DCC_WHITE_CONTINUE:
		case DCC_WHITE_COMPLAIN:
			cmn_wf.changed = 1;
			dcc_error_msg("%s", emsg);
			break;
		}
	}

	totals_init();
}



void
lock_work(void)
{
	int result = pthread_mutex_lock(&work_mutex);
	if (result)
		dcc_logbad(EX_SOFTWARE, "pthread_mutex_lock(work_free): %s",
			   ERROR_STR1(result));
}



void
unlock_work(void)
{
	int result = pthread_mutex_unlock(&work_mutex);
	if (result)
		dcc_logbad(EX_SOFTWARE, "pthread_mutex_unlock(work_free): %s",
			   ERROR_STR1(result));
}



static void
wf_lock(void)
{
	int error = pthread_mutex_lock(&cwf_mutex);
	if (error)
		dcc_logbad(EX_SOFTWARE, "pthread_mutex_lock(cwf): %s",
			   ERROR_STR1(error));
}



static void
wf_unlock(void)
{
	int error = pthread_mutex_unlock(&cwf_mutex);
	if (error)
		dcc_logbad(EX_SOFTWARE, "pthread_mutex_unlock(cwf): %s",
			   ERROR_STR1(error));
}



void
cmn_create(CMN_WORK *cwp)
{
	cwp->tmp_fd = -1;
	cwp->log_fd = -1;
	cwp->log_fd2 = -1;
}



u_char
cmn_open_tmp(CMN_WORK *cwp, u_char anonymous)
{
	if (!tmpdir)
		tmpdir = dcc_logdir[0] != '\0' ? dcc_logdir : _PATH_TMP;

	if (!anonymous)
		cwp->cmn_fgs |= CMN_FG_TMP_UNLINK;

	cwp->tmp_fd = dcc_mkstemp(cwp->emsg,
				  cwp->tmp_nm, sizeof(cwp->tmp_nm),
				  cwp->id, sizeof(cwp->id),
				  tmpdir, DCC_TMP_LOG_PAT,
				  anonymous, 0, 0);

	if (cwp->tmp_fd < 0 && tmpdir == dcc_logdir) {
		cwp->tmp_fd = dcc_mkstemp(cwp->emsg,
					  cwp->tmp_nm, sizeof(cwp->tmp_nm),
					  cwp->id, sizeof(cwp->id),
					  _PATH_TMP, DCC_TMP_LOG_PAT,
					  anonymous, 0, 0);
	}

	return cwp->tmp_fd >= 0;
}



void
cmn_close_tmp(CMN_WORK *cwp)
{
	if (cwp->tmp_fd >= 0) {
		if (0 > close(cwp->tmp_fd))
			thr_error_msg(cwp, "close(%s): %s",
				      cwp->tmp_nm, ERROR_STR());
		cwp->tmp_fd = -1;
		if (cwp->cmn_fgs & CMN_FG_TMP_UNLINK) {
			if (0 < unlink(cwp->tmp_nm))
				dcc_error_msg("unlink(%s): %s",
					      cwp->tmp_nm, ERROR_STR());
			cwp->cmn_fgs &= ~CMN_FG_TMP_UNLINK;
		}
	}
	cwp->tmp_nm[0] = '\0';
}



u_char
cmn_write_tmp(CMN_WORK *cwp, const void *buf, int len)
{
	int i;

	if (cwp->tmp_fd < 0)
		return 1;

	i = write(cwp->tmp_fd, buf, len);
	if (i == len)
		return 1;

	if (i < 0)
		thr_error_msg(cwp, "write(%s,%d): %s",
			      cwp->tmp_nm, len, ERROR_STR());
	else
		thr_error_msg(cwp, "write(%s,%d)=%d",
			      cwp->tmp_nm, len, i);
	cmn_close_tmp(cwp);
	return 0;
}



void
cmn_clear(CMN_WORK *cwp, struct work *wp)
{
	log_stop(cwp);

	cmn_close_tmp(cwp);

	if (cwp->num_rcpts)
		rcpts_free(cwp, 1);

	cwp->cmn_fgs = 0;
	cwp->emsg[0] = '\0';
	cwp->id[0] = '\0';
	cwp->env_from[0] = '\0';
	cwp->early_log.len = 0;
	memset(&cwp->CMN_WORK_ZERO, 0,
	       sizeof(*cwp) - ((char*)&cwp->CMN_WORK_ZERO - (char*)cwp));
	cwp->wp = wp;
	cwp->action = action;
}



void
rcpts_create(int i)
{
	RCPT_ST *rcpt_st;

	rcpt_st = dcc_malloc(sizeof(*rcpt_st)*i);
	rcpt_st_free = rcpt_st;
	memset(rcpt_st, 0, sizeof(*rcpt_st)*i);
	while (--i > 0) {
		rcpt_st->fwd = rcpt_st+1;
		rcpt_st->user_log_fd = -1;
		++rcpt_st;
	}
}



void
rcpts_free(CMN_WORK *cwp, u_char need_lock)
{
	RCPT_ST *rcpt_st, *next_rcpt_st;

	rcpt_st = cwp->rcpt_st_first;
	if (!rcpt_st)
		return;

	if (need_lock)
		lock_work();
	cwp->rcpt_st_first = 0;
	do {
		next_rcpt_st = rcpt_st->fwd;
		if (rcpt_st->user_log_fd >= 0) {
			if (0 > close(rcpt_st->user_log_fd))
				dcc_error_msg("close(%s): %s",
					      rcpt_st->user_log_nm,
					      ERROR_STR());
			rcpt_st->user_log_fd = -1;
		}
		rcpt_st->fwd = rcpt_st_free;
		rcpt_st_free = rcpt_st;
	} while ((rcpt_st = next_rcpt_st) != 0);
	cwp->num_rcpts = 0;

	if (need_lock)
		unlock_work();
}



RCPT_ST *
rcpt_st_alloc(CMN_WORK *cwp,
	   u_char unlocked)		/* 1=unlocked on entry & exit */
{
	RCPT_ST *rcpt_st;

	if (cwp->num_rcpts >= MAX_RCPTS) {
		thr_error_msg(cwp, "too many recipients");
		return 0;
	}

	if (unlocked)
		lock_work();
	rcpt_st = rcpt_st_free;
	if (rcpt_st) {
		rcpt_st_free = rcpt_st->fwd;
		if (unlocked)
			unlock_work();
	} else {
		if (unlocked)
			unlock_work();
		rcpt_st = dcc_malloc(sizeof(*rcpt_st));
		memset(rcpt_st, 0, sizeof(*rcpt_st));
		rcpt_st->user_log_fd = -1;
	}
	rcpt_st->fwd = 0;
	rcpt_st->log_pos_white = 0;
	memset(rcpt_st->wtgts, 0, sizeof(rcpt_st->wtgts));
	rcpt_st->embargo_num = 0;
	rcpt_st->flags = 0;
	rcpt_st->user[0] = '\0';
	rcpt_st->dir[0] = '\0';
	rcpt_st->user_log_nm[0] = '\0';

	/* note position in the main log file so that we can later pick out
	 * this recipient's part of the envelope */
	rcpt_st->cwp = cwp;
	if (!cwp->rcpt_st_first) {
		cwp->rcpt_st_first = rcpt_st;
		cwp->log_pos_to_first = log_lseek(cwp, SEEK_END);
		rcpt_st->log_pos_to = cwp->log_pos_to_first;
	} else {
		cwp->rcpt_st_last->fwd = rcpt_st;
		rcpt_st->log_pos_to = cwp->log_pos_to_end;
	}
	cwp->rcpt_st_last = rcpt_st;
	++cwp->num_rcpts;

	return rcpt_st;
}



void
parse_userdirs(const char *arg)
{
	static DCC_PATH userdirs_path;

	/* add '/' to end of the path without converting it to "/" */
	if (*arg == '\0') {
		userdirs_path[0] = '\0';
		userdirs_len = 0;
	} else {
		strncpy(userdirs_path, arg,
			sizeof(userdirs_path));
		userdirs_len = strlen(userdirs_path)-1;
		while (userdirs_len > 1 && userdirs_path[userdirs_len] == '/')
			--userdirs_len;
		userdirs_path[++userdirs_len] = '/';
		userdirs_path[++userdirs_len] = '\0';

	}
	userdirs = userdirs_path;
}



/* sanitize recipient mailbox and per-user log and whitelist directory */
u_char					/* 0=complain about something */
get_user_dir(RCPT_ST *rcpt_st,
	     const char *str1, int str1_len, const char *str2, int str2_len)
{
	char *p;
	char c;
	u_char seen_slash;
	int dots;
	int i;

	if (!userdirs) {
		rcpt_st->dir[0] = '\0';
		return 1;
	}

	memcpy(rcpt_st->dir, userdirs, userdirs_len);
	i = userdirs_len;
	if (i+str1_len < ISZ(rcpt_st->dir))
		memcpy(&rcpt_st->dir[i], str1, str1_len);
	i += str1_len;
	if (str2) {
		if (i+str2_len+1 < ISZ(rcpt_st->dir)) {
			rcpt_st->dir[i++] = '/';
			memcpy(&rcpt_st->dir[i], str2, str2_len);
		}
		i += str2_len;
	}
	if (i >= ISZ(rcpt_st->dir)) {
		dcc_pemsg(EX_DATAERR, rcpt_st->cwp->emsg,
			  "recipient \"%s\" is too long", rcpt_st->dir);
		rcpt_st->dir[0] = '\0';
		return 0;
	}
	rcpt_st->dir[i] = '\0';

	/* To get a consistent directory name,
	 * convert ASCII upper to lower case.
	 * Be simplistic about international character sets and
	 * avoid locale and portability complications.
	 * Refuse insecure paths. */
	seen_slash = 1;			/* userdirs ends with '/' */
	dots = 0;
	p = &rcpt_st->dir[userdirs_len];
	for (;;) {
		c = *p;
		if (c == '/' || c == '\\' || c == '\0') {
			if (dots == 2) {
				dcc_pemsg(EX_DATAERR, rcpt_st->cwp->emsg,
					  "path \"%s\" is insecure",
					  rcpt_st->dir);
				rcpt_st->dir[0] = '\0';
				return 0;
			}
			if (c == '\0')
				break;
			seen_slash = 1;
			dots = 0;
		} else if (c == '.' && seen_slash && dots <= 1) {
			++dots;
		} else {
			*p = DCC_TO_LOWER(c);
			seen_slash = 0;
			dots = 0;
		}
		++p;
	}

	return 1;
}



/* start main log file */
u_char					/* 1=new file, 0=old file or failed */
log_start(CMN_WORK *cwp)
{
	struct tm tm;
	char date_buf[40];

	if (dcc_logdir[0] == '\0')
		return 0;
	if (cwp->log_fd >= 0)
		return 0;

#if MAX_LOG_SIZE > 0
	cwp->log_size = 0;
#endif
	cwp->log_fd = dcc_log_open(cwp->emsg, cwp->log_nm, sizeof(cwp->log_nm),
				   cwp->id, sizeof(cwp->id));
	if (cwp->log_fd < 0) {
		static time_t whined;
		time_t now;

		now = time(0);
		if (now < whined || now > whined+5*60 || dcc_clnt_debug)
			dcc_error_msg("%s", cwp->emsg);
		whined = now;
		cwp->emsg[0] = '\0';
		return 0;
	}

	cwp->cmn_fgs |= CMN_FG_LOG_EARLY;

	strftime(date_buf, sizeof(date_buf), DCC_LOG_DATE_FMT,
		 dcc_localtime(time(0), &tm));
	thr_log_print(cwp, 0, DCC_LOG_DATE_PAT"\n", date_buf);
	return 1;
}



static void
log_fd2_close(CMN_WORK *cwp, int flag)
{
	if (cwp->log_fd2 >= 0
	    && 0 > close(cwp->log_fd2))
		dcc_error_msg("close(FD2 %s): %s",
			      cwp->log_nm, ERROR_STR());
	cwp->log_fd2 = flag;
}



void
log_stop(CMN_WORK *cwp)
{
	thr_log_late(cwp);
	log_fd2_close(cwp, -1);

	if (cwp->log_fd < 0)
		return;

	/* Close before renaming to accomodate WIN32 foolishness.
	 * Assuming dcc_mkstemp() works properly, there is no race */
	if (0 > close(cwp->log_fd))
		dcc_error_msg("close(%s): %s", cwp->log_nm, ERROR_STR());
	if (!(cwp->ask_st & (ASK_ST_LOGIT
			     | ASK_ST_LOG_ONE
			     | ASK_ST_GREY_LOGIT))) {
		/* Delete the log file if it is not interesting */
		unlink(cwp->log_nm);
	} else {
		/* give it a permanent name if it is interesting */
		dcc_log_keep(0, cwp->log_nm, sizeof(cwp->log_nm));
	}
	cwp->log_nm[0] = '\0';
	cwp->log_fd = -1;
}



void
log_write(CMN_WORK *cwp, const void *logbuf, u_int len)
{
	int result;

	if (cwp->log_fd < 0)
		return;

	if (!len)
		len = strlen(logbuf);
#if MAX_LOG_SIZE > 0
	cwp->log_size += len;
#endif

	result = write(cwp->log_fd, logbuf, len);
	if ((int)len == result)
		return;

	if (result < 0)
		dcc_error_msg("write(%s): %s", cwp->log_nm, ERROR_STR());
	else
		dcc_error_msg("write(%s,%d)=%d", cwp->log_nm, len, result);
	close(cwp->log_fd);
	cwp->log_fd = -1;
}



static inline void
log_ck_write(void *cwp, const void *logbuf, u_int len)
{
	log_write((CMN_WORK *)cwp, logbuf, len);
}



off_t
log_lseek(CMN_WORK *cwp, int whence)
{
	off_t result;

	if (cwp->log_fd < 0)
		return 0;
	result = lseek(cwp->log_fd, 0, whence);
	if (result == -1) {
		thr_error_msg(cwp, "lseek(%s, 0, %d): %s",
			      cwp->log_nm, whence, ERROR_STR());
		close(cwp->log_fd);
		cwp->log_fd = -1;
		return 0;
	}
	return result;
}



/* put something into a log file */
static void
vthr_log_print(CMN_WORK *cwp, u_char error, const char *p, va_list args)
{
	char logbuf[MAXHOSTNAMELEN*3];

	if (cwp->log_fd < 0
	    || ((cwp->cmn_fgs & CMN_FG_LOG_EARLY)
		&& error)) {
		dcc_vearly_log(&cwp->early_log, p, args);
		return;
	}

	vsnprintf(logbuf, sizeof(logbuf), p, args);
	log_write(cwp, logbuf, 0);
}



void PATTRIB(3,4)
thr_log_print(void *cwp, u_char error, const char *p, ...)
{
	va_list args;

	va_start(args, p);
	vthr_log_print(cwp, error, p, args);
	va_end(args);
}



void
thr_error_msg(void *cwp0, const char *p, ...)
{
	CMN_WORK *cwp = cwp0;
	va_list args;

	va_start(args, p);
	dcc_verror_msg(p, args);
	va_end(args);

	va_start(args, p);
	vthr_log_print(cwp, 1, p, args);
	va_end(args);
	thr_log_print(cwp, 1, "\n");

	cwp->ask_st |= ASK_ST_LOGIT;
}



void
thr_trace_msg(void *cwp0, const char *p, ...)
{
	va_list args;

	va_start(args, p);
	dcc_vtrace_msg(p, args);
	va_end(args);

	if (cwp0) {
		CMN_WORK *cwp = cwp0;
		va_start(args, p);
		vthr_log_print(cwp, 1, p, args);
		va_end(args);
		thr_log_print(cwp, 1, "\n");

		cwp->ask_st |= ASK_ST_LOGIT;
	}
}



void
thr_log_late(CMN_WORK *cwp)
{
	cwp->cmn_fgs &= ~CMN_FG_LOG_EARLY;
	if (cwp->early_log.len) {
		log_write(cwp, cwp->early_log.buf, cwp->early_log.len);
		cwp->early_log.len = 0;
	}
}



/* open the connection to the nearest DCC server */
u_char
ck_dcc_ctxt(CMN_WORK *cwp)
{
	if (cwp->dcc_ctxt_sn != dcc_ctxt_sn) {
		cwp->dcc_ctxt_sn = dcc_ctxt_sn;
		cwp->dcc_ctxt = dcc_clnt_init(cwp->emsg, cwp->dcc_ctxt, 0,
					      DCC_CLNT_FG_NO_SRVR_OK
					      | DCC_CLNT_FG_NO_FAIL);
		if (!cwp->dcc_ctxt) {
			/* failed to create context */
			thr_error_msg(cwp, "%s", cwp->emsg);
			cwp->dcc_ctxt_sn = 0;
			return 0;
		}
		cwp->xhdr_len = dcc_xhdr_start(cwp->xhdr, sizeof(cwp->xhdr),
					       dcc_clnt_info);
	}
	return 1;
}



/* find and lock a DCC_WF */
static CWF *
find_cwf(RCPT_ST *rcpt_st)
{
	CWF *cwf;
	DCC_PATH white_nm0, white_nm;
	const char *white_nm_ptr;
	struct stat sb;

	if (rcpt_st->dir[0] == '\0')
		return 0;

	/* use the full path for the key */
	snprintf(white_nm0, sizeof(white_nm0), "%s/whiteclnt", rcpt_st->dir);
	dcc_fnm2path(white_nm, white_nm0);

	wf_lock();
	white_nm_ptr = white_nm;
	cwf = cur_cwf;
	for (;;) {
		if (!strcmp(white_nm_ptr, cwf->wf.ascii_nm)) {
			/* we found a DCC_WF for the target file */
			white_nm_ptr = 0;
			break;
		}

		if (cwf->older == cur_cwf) {
			/* we do not know this whitefile
			 * If neither the ASCII or hash table files exist,
			 * quite now.
			 * Otherwise recycle the oldest DCC_WF
			 *
			 * We are in the home directory and so can
			 * use the short name and save some work
			 * for the kernel */
			if (0 <= stat(white_nm0, &sb))
				break;
			if (errno != ENOENT) {
				if (errno != ENOTDIR || dcc_clnt_debug > 2)
					thr_error_msg(rcpt_st->cwp,
						      "stat(%s): %s",
						      white_nm, ERROR_STR());
				wf_unlock();
				return 0;
			}
			snprintf(white_nm0, sizeof(white_nm0),
				 "%s/whiteclnt"DCC_WHITE_SUFFIX,
				 rcpt_st->dir);
			if (0 <= stat(white_nm0, &sb))
				break;
			if (errno != ENOENT
			    && (errno != ENOTDIR || dcc_clnt_debug))
				thr_error_msg(rcpt_st->cwp, "stat(%s): %s",
					      dcc_fnm2path(white_nm, white_nm0),
					      ERROR_STR());
			wf_unlock();
			return 0;
		}

		cwf = cwf->older;
	}

	/* move to front */
	if (cwf != cur_cwf) {
		cwf->older->newer = cwf->newer;
		cwf->newer->older = cwf->older;
		cwf->older = cur_cwf;
		cwf->newer = cur_cwf->newer;
		cwf->newer->older = cwf;
		cwf->older->newer = cwf;
		cur_cwf = cwf;
	}

	if (white_nm_ptr
	    && !dcc_new_white_nm(rcpt_st->cwp->emsg, &cwf->wf, white_nm_ptr)) {
		thr_error_msg(rcpt_st->cwp, "%s", rcpt_st->cwp->emsg);
		wf_unlock();
		return 0;
	}

	switch (dcc_rdy_white(rcpt_st->cwp->emsg, &cwf->wf)) {
	case DCC_WHITE_OK:
	case DCC_WHITE_NOFILE:
		break;
	case DCC_WHITE_SILENT:
		if (dcc_clnt_debug)
			thr_error_msg(rcpt_st->cwp, "%s", rcpt_st->cwp->emsg);
		break;
	case DCC_WHITE_CONTINUE:
		thr_error_msg(rcpt_st->cwp, "%s", rcpt_st->cwp->emsg);
		break;
	case DCC_WHITE_COMPLAIN:
		wf_unlock();
		thr_error_msg(rcpt_st->cwp, "%s", rcpt_st->cwp->emsg);
		return 0;
	}

	return cwf;
}



/* digest the results of a whitelist */
static void
white_results(RCPT_ST *rcpt_st,
	      CMN_WORK *cwp,
	      RCPT_ST_FLAGS *flagsp,
	      DCC_WHITE_RESULT result,
	      DCC_WHITE_LISTING *listingp)
{

	DCC_WHITE_LISTING listing;

	/* call-by-reference parameter to resolve order of evaluation
	 * in callers */
	listing = *listingp;

	switch (result) {
	case DCC_WHITE_OK:
		break;
	case DCC_WHITE_SILENT:
	case DCC_WHITE_NOFILE:
		listing = DCC_WHITE_UNLISTED;
		break;
	case DCC_WHITE_COMPLAIN:
	case DCC_WHITE_CONTINUE:
		thr_error_msg(cwp, "%s", cwp->emsg);
		listing = DCC_WHITE_UNLISTED;
	}

	switch (listing) {
	case DCC_WHITE_HALF_LISTED:
		/* "OK2" for the RCPT in the local whitelist does not mean
		 * the address is half whitelisted, but that it is
		 * explicitly ok to reject or discard spam for it based
		 * on DCC results.
		 * We get this value only from dcc_white_sum() */
		if (rcpt_st)
			rcpt_st->flags &= ~RCPT_ST_DCC_OFF;
		break;

	case DCC_WHITE_LISTED:
		*flagsp |= RCPT_ST_WHITE;
		*flagsp &= ~RCPT_ST_BLACK;
		break;

	case DCC_WHITE_UNLISTED:
		break;

	case DCC_WHITE_BLACK:
		if (!(*flagsp & RCPT_ST_WHITE))
			*flagsp |= RCPT_ST_BLACK;
		break;
	}
}



static void
want_grey_on(const DCC_WF *wf)
{
	static u_char once;
	DCC_PATH abs_nm;

	if (once)
		return;
	once = 1;

	if (!grey_on)
		dcc_error_msg("%s wants greylisting but it is turned off",
			      dcc_fnm2path(abs_nm, wf->ascii_nm));
}



static void
rcpt_st_flags2ask_st(CMN_WORK *cwp, RCPT_ST_FLAGS flags)
{
	/* it is spam if it is blacklisted by any target */
	if (flags & RCPT_ST_BLACK)
		cwp->ask_st |= (ASK_ST_LOCAL_WASSPAM | ASK_ST_LOGIT);

	/* If we had a DNS blacklist hit on a URL
	 * and this recipient believes the DNS blacklists,
	 * then it is spam to report to DCC server */
	if ((flags & RCPT_ST_DNSBL_ON)
	    && (cwp->ask_st & ASK_ST_DNSBL_ISSPAM))
		cwp->ask_st |= (ASK_ST_LOCAL_WASSPAM | ASK_ST_LOGIT);

	/* If the external flter said this is spam
	 * and this recipient cares,
	 * then it is spam to report to the DCC server */
	if ((flags & RCPT_ST_XFLTR_ON)
	    && (cwp->ask_st & ASK_ST_XFLTR_ISSPAM))
		cwp->ask_st |= (ASK_ST_LOCAL_WASSPAM | ASK_ST_LOGIT);
}



static void
wf2rcpt_st_flags(RCPT_ST_FLAGS *flagsp, const DCC_WF *wf)
{
	if ((wf->info_flags & DCC_WHITE_FG_DCC_OFF))
		*flagsp |= RCPT_ST_DCC_OFF;
	else if (wf->info_flags & DCC_WHITE_FG_DCC_ON)
		*flagsp &= ~RCPT_ST_DCC_OFF;

	if (wf->info_flags & DCC_WHITE_FG_GREY_ON) {
		*flagsp &= ~RCPT_ST_GREY_OFF;
		want_grey_on(wf);
	}
	if (!grey_on || (wf->info_flags & DCC_WHITE_FG_GREY_OFF))
		*flagsp |= RCPT_ST_GREY_OFF;

	if (wf->info_flags & DCC_WHITE_FG_GREY_LOG_ON) {
		*flagsp &= ~RCPT_ST_GREY_LOG_OFF;
		want_grey_on(wf);
	} else if (wf->info_flags & DCC_WHITE_FG_GREY_LOG_OFF) {
		*flagsp |= RCPT_ST_GREY_LOG_OFF;
	}

	if (wf->info_flags & DCC_WHITE_FG_LOG_ALL) {
		*flagsp |= RCPT_ST_LOG_ALL;
	} else if (wf->info_flags & DCC_WHITE_FG_LOG_NORMAL) {
		*flagsp &= ~RCPT_ST_LOG_ALL;
	}

	if (wf->info_flags & DCC_WHITE_FG_MTA_FIRST) {
		*flagsp |= RCPT_ST_MTA_FIRST;
	} else if (wf->info_flags & DCC_WHITE_FG_MTA_LAST) {
		*flagsp &= ~RCPT_ST_MTA_FIRST;
	}

	if ((wf->info_flags & DCC_WHITE_FG_DNSBL_ON)
	    && dnsbls) {
		*flagsp |= RCPT_ST_DNSBL_ON;
	} else if (wf->info_flags & DCC_WHITE_FG_DNSBL_OFF) {
		*flagsp &= ~RCPT_ST_DNSBL_ON;
	}

#ifdef USE_XFLTR
	if ((wf->info_flags & DCC_WHITE_FG_XFLTR_ON)
	    && xfltr_parm) {
		*flagsp |= RCPT_ST_XFLTR_ON;
	} else if (wf->info_flags & DCC_WHITE_FG_XFLTR_OFF) {
		*flagsp &= ~RCPT_ST_XFLTR_ON;
	}
#endif
}



/* check the whitelists for a single user or target */
static void
ask_white_rcpt(CMN_WORK *cwp,
	       RCPT_ST *rcpt_st,
	       RCPT_ST_FLAGS global_flags)
{
	DCC_SUM addr_sum;
	DCC_WHITE_LISTING listing;
	CWF *cwf;
	DCC_PATH abs_nm;

	rcpt_st->log_pos_white = log_lseek(cwp, SEEK_END);
	rcpt_st->flags |= cwp->rcpt_st_flags & (RCPT_ST_DCC_OFF
						| RCPT_ST_DCC_REP_ON
						| RCPT_ST_GREY_OFF
						| RCPT_ST_GREY_LOG_OFF
						| RCPT_ST_MTA_FIRST
						| RCPT_ST_XFLTR_ON
						| RCPT_ST_DNSBL_ON
						| RCPT_ST_LOG_ALL);

	/* per-user whitelist options overrides global whitelist */
	cwf = find_cwf(rcpt_st);
	if (cwf)
		wf2rcpt_st_flags(&rcpt_st->flags, &cwf->wf);

	/* Compute white- or blacklisting.
	 *	The per-user whiteclnt file overrides the global file.
	 *	The whiteclnt files override the MTA with "option MTA-first".
	 *	    Without, the MTA controls.
	 *	Within each category, whitelisting overrides blacklisting.
	 *
	 *	The global whitelist's answer for message's checksums (other
	 *	    than the env_to checksums) have already been computed in
	 *	    global_flags. */

	if (rcpt_st->flags & RCPT_ST_MTA_FIRST) {
		if (cwp->ask_st & ASK_ST_MTA_NOTSPAM) {
			rcpt_st->flags |= RCPT_ST_WHITE;
			rcpt_st->flags &= ~RCPT_ST_BLACK;
		} else if (cwp->ask_st & ASK_ST_MTA_ISSPAM) {
			rcpt_st->flags |= RCPT_ST_BLACK;
			rcpt_st->flags &= ~RCPT_ST_WHITE;
		}
	}

	dcc_str2ck(rcpt_st->env_to_sum, 0, 0, rcpt_st->env_to);
	if (rcpt_st->user[0])
		dcc_str2ck(addr_sum, 0, 0, rcpt_st->user);


	/* check the global whitelist for env_to & mailbox because we
	 * might need the result and to check for DCC_WHITE_HALF_LISTED */
	if (!cwf)
		wf_lock();
	white_results(rcpt_st, cwp, &global_flags,
		      dcc_white_sum(cwp->emsg, &cmn_wf,
				    DCC_CK_ENV_TO,
				    rcpt_st->env_to_sum, &listing),
		      &listing);
	/* check the mailbox name (after aliases etc.) */
	if (rcpt_st->user[0] != '\0')
		white_results(rcpt_st, cwp, &global_flags,
			      dcc_white_sum(cwp->emsg, &cmn_wf,
					    DCC_CK_ENV_TO,
					    addr_sum, &listing),
			      &listing);
	if (!cwf)
		wf_unlock();

	if (global_flags != 0) {
		rcpt_st->flags &= ~(RCPT_ST_WHITE | RCPT_ST_BLACK);
		rcpt_st->flags |= global_flags;
	}

	/* check the per-user whitelist */
	if (cwf) {
		RCPT_ST_FLAGS flags = 0;

		/* check the envelope Rcpt_To value */
		white_results(rcpt_st, cwp, &flags,
			      dcc_white_sum(cwp->emsg, &cwf->wf,
					    DCC_CK_ENV_TO,
					    rcpt_st->env_to_sum, &listing),
			      &listing);

		/* check the mailbox name (after aliases etc.) */
		if (rcpt_st->user[0] != '\0')
			white_results(rcpt_st, cwp, &flags,
				      dcc_white_sum(cwp->emsg, &cwf->wf,
						    DCC_CK_ENV_TO,
						    addr_sum, &listing),
				      &listing);
		white_results(rcpt_st, cwp, &flags,
			      dcc_white_cks(cwp->emsg, &cwf->wf,
					    &cwp->cks,
					    rcpt_st->wtgts, &listing),
			      &listing);

		if (flags != 0) {
			thr_log_print(cwp, 0, "%s%s\n",
				      dcc_fnm2path(abs_nm, cwf->wf.ascii_nm),
				      (flags & RCPT_ST_WHITE)
				      ? DCC_XHDR_ISOK
				      : DCC_XHDR_ISSPAM);
			rcpt_st->flags &= ~(RCPT_ST_WHITE | RCPT_ST_BLACK);
			rcpt_st->flags |= flags;
		}

		wf_unlock();
	}

	if (!(rcpt_st->flags & RCPT_ST_MTA_FIRST)) {
		if (cwp->ask_st & ASK_ST_MTA_NOTSPAM) {
			rcpt_st->flags |= RCPT_ST_WHITE;
			rcpt_st->flags &= ~RCPT_ST_BLACK;
		} else if (cwp->ask_st & ASK_ST_MTA_ISSPAM) {
			rcpt_st->flags |= RCPT_ST_BLACK;
			rcpt_st->flags &= ~RCPT_ST_WHITE;
		}
	}
}



/* check the whitelists for all targets */
void
cmn_ask_white(CMN_WORK *cwp, u_char mta_grey_query)
{
	RCPT_ST *rcpt_st;
	RCPT_ST_FLAGS global_flags;
	DCC_OPS grey_op;
	DCC_WHITE_LISTING listing;

	dcc_dnsbl_result(&cwp->ask_st, cwp->cks.dnsbl);

#ifdef USE_XFLTR
	if (cwp->tmp_fd >= 0
	    && xfltr_parm) {
		if (ask_xfltr(cwp->dcc_ctxt, cwp,
			      cwp->xfltr_header, sizeof(cwp->xfltr_header),
			      cwp->id,
			      cwp->clnt_name, &cwp->clnt_addr,
			      cwp->helo, cwp->env_from, cwp->num_rcpts,
			      cwp->tmp_nm)) {
			cwp->ask_st |= (ASK_ST_XFLTR_ISSPAM | ASK_ST_LOGIT);
		}
	}
#endif

	cwp->log_pos_white_first = log_lseek(cwp, SEEK_END);

	/* Use the main whitelist only for recipients whose individual
	 * whitelists don't give a black or white answer.
	 * Check the main whitelist first (and so even if not necessary)
	 * so that problems with it are in all of the logs and to simplify
	 * merging the global and per-user whitelist results. */
	wf_lock();
	global_flags = 0;
	white_results(0, cwp, &global_flags,
		      dcc_white_cks(cwp->emsg, &cmn_wf, &cwp->cks,
				    cwp->wtgts, &listing),
		      &listing);

	/* get the defaults for the options */
	if (to_white_only)
		cwp->rcpt_st_flags |= RCPT_ST_DCC_OFF;
	wf2rcpt_st_flags(&cwp->rcpt_st_flags, &cmn_wf);
	wf_unlock();

	/* special case kludge for no recipients for dccifd
	 *	leave cwp->white_tgts == 1 and != cwp->tgts
	 *	if the message is not whitelist so that dccifd will
	 *	ask the DCC server
	 */
	rcpt_st = cwp->rcpt_st_first;
	if (!rcpt_st) {
		if (cwp->rcpt_st_flags & RCPT_ST_MTA_FIRST) {
			if (cwp->ask_st & ASK_ST_MTA_NOTSPAM) {
				cwp->rcpt_st_flags |= RCPT_ST_WHITE;
				cwp->rcpt_st_flags &= ~RCPT_ST_BLACK;
			} else if (cwp->ask_st & ASK_ST_MTA_ISSPAM) {
				cwp->rcpt_st_flags |= RCPT_ST_BLACK;
				cwp->rcpt_st_flags &= ~RCPT_ST_WHITE;
			}
		}
		if (global_flags != 0) {
			cwp->rcpt_st_flags &= ~(RCPT_ST_WHITE | RCPT_ST_BLACK);
			cwp->rcpt_st_flags |= global_flags;
		}
		if (!(cwp->rcpt_st_flags & RCPT_ST_MTA_FIRST)) {
			if (cwp->ask_st & ASK_ST_MTA_NOTSPAM) {
				cwp->rcpt_st_flags |= RCPT_ST_WHITE;
				cwp->rcpt_st_flags &= ~RCPT_ST_BLACK;
			} else if (cwp->ask_st & ASK_ST_MTA_ISSPAM) {
				cwp->rcpt_st_flags |= RCPT_ST_BLACK;
				cwp->rcpt_st_flags &= ~RCPT_ST_WHITE;
			}
		}
		if (!(cwp->rcpt_st_flags & RCPT_ST_WHITE))
			++cwp->white_tgts;
		rcpt_st_flags2ask_st(cwp, cwp->rcpt_st_flags);
	}

	for (; rcpt_st; rcpt_st = rcpt_st->fwd) {
		/* maybe this recipient is whitelisted or a spam trap
		 * or has per-user option settings */
		ask_white_rcpt(cwp, rcpt_st, global_flags);

		/* We need to know if all targets are whitelisted for the DCC
		 * before we ask the DCC server.  Mail sent only to whitelisted
		 * targets should not be reported to the DCC server.
		 * For that we need a count of white-listed targets */
		if (rcpt_st->flags & RCPT_ST_WHITE) {
			++cwp->white_tgts;
		} else {
			rcpt_st_flags2ask_st(cwp, rcpt_st->flags);
		}

		if (rcpt_st->flags & RCPT_ST_GREY_OFF) {
			rcpt_st->grey_result = ASK_GREY_OFF;
			continue;
		}

		if ((rcpt_st->flags & RCPT_ST_BLACK)
		    || grey_query_only
		    || mta_grey_query)
			grey_op = DCC_OP_GREY_QUERY;
		else if (rcpt_st->flags & RCPT_ST_WHITE)
			grey_op = DCC_OP_GREY_WHITE;
		else
			grey_op = DCC_OP_GREY_REPORT;
		rcpt_st->grey_result = ask_grey(cwp->emsg, cwp->dcc_ctxt,
						grey_op,
						rcpt_st->msg_sum,
						rcpt_st->triple_sum,
						&cwp->cks,
						rcpt_st->env_to_sum,
						&rcpt_st->embargo_num,
						&cwp->early_grey_tgts,
						&cwp->late_grey_tgts);

		switch (rcpt_st->grey_result) {
		case ASK_GREY_OFF:
		case ASK_GREY_SPAM:
			break;		/* impossible */
		case ASK_GREY_FAIL:
			thr_error_msg(cwp, "%s", cwp->emsg);
			/* If we are trying hard, assume the
			 * message would have been embargoed */
			if (try_extra_hard)
				cwp->ask_st |= (ASK_ST_GREY_EMBARGO
						| ASK_ST_GREY_LOGIT);
			break;
		case ASK_GREY_EMBARGO:
			if (rcpt_st->embargo_num == 1
			    && (rcpt_st->flags & RCPT_ST_BLACK)) {
				/* don't bother revoking non-existent entry */
				rcpt_st->grey_result = ASK_GREY_OFF;
			} else {
				cwp->ask_st |= (ASK_ST_GREY_EMBARGO
						| ASK_ST_GREY_LOGIT);
			}
			break;
		case ASK_GREY_EMBARGO_END:
			cwp->ask_st |= ASK_ST_GREY_LOGIT;
			rcpt_st->flags |= RCPT_ST_GREY_END;
			break;
		case ASK_GREY_PASS:
			break;
		case ASK_GREY_WHITE:
			rcpt_st->flags |= RCPT_ST_GREY_WHITE;
			break;
		}
	}

	cwp->log_pos_white_last = log_lseek(cwp, SEEK_END);
}



/* ask a DCC server */
int					/* <0=big problem, 0=retryable, 1=ok */
cmn_ask_dcc(CMN_WORK *cwp, u_char mta_query)
{
	int i;
	u_char spam;

	/* Talk to the DCC server and make the X-DCC header.
	 * If we have blacklist entries for it, then we'll tell the DCC
	 * server it is spam and say so in the X-DCC header.
	 * Note that a target count of 0 is a query. */
	if (dcc_query_only || mta_query) {
		spam = 0;
	} else if (cwp->ask_st & ASK_ST_LOCAL_WASSPAM) {
		spam = 1;
		cwp->local_tgts = cwp->tgts;
	} else if (cwp->ask_st & ASK_ST_GREY_EMBARGO) {
		/* if the message is under a greylist embargo,
		 * then report to the DCC only the targets for
		 * which it is an initial transmission or embargo #1 */
		spam = 0;
		cwp->local_tgts = cwp->early_grey_tgts;
	} else {
		/* if this is the end of a greylist embargo
		 * then do not tell the DCC about targets that were
		 * counted with previous transmissions.  Those targets
		 * are counted in late_grey_tgts this time, but were
		 * counted in early_grey_tgts for previous transmissions */
		spam = 0;
		cwp->local_tgts = cwp->tgts - cwp->late_grey_tgts;
	}
	i = ask_dcc(cwp->emsg, cwp->dcc_ctxt, try_extra_hard,
		    &cwp->header, &cwp->cks, &cwp->ask_st, spam,
		    cwp->local_tgts);
	if (i <= 0) {
		thr_error_msg(cwp, "%s", cwp->emsg);
		return i;
	}

	/* if we are talking to a new server,
	 * remember to fix the X-DCC headers of the other contexts */
	if (cwp->xhdr_len != cwp->header.start_len
	    || strncmp(cwp->header.buf, cwp->xhdr, cwp->xhdr_len)) {
		if (dcc_clnt_debug)
			thr_trace_msg(cwp, DCC_XHDR_START
				      "header changed from %s to %.*s",
				      cwp->xhdr,
				      (int)cwp->header.start_len,
				      cwp->header.buf);
		cwp->xhdr_len = dcc_xhdr_start(cwp->xhdr, sizeof(cwp->xhdr),
					       dcc_clnt_info);
		cwp->dcc_ctxt_sn = ++dcc_ctxt_sn;
	}

	return 1;
}



u_char
user_log_write(RCPT_ST *rcpt_st, const void *buf, u_int len)
{
	DCC_PATH abs_nm;
	int result;

	if (rcpt_st->user_log_fd < 0)
		return 0;

	if (!len)
		len = strlen(buf);
	result = write(rcpt_st->user_log_fd, buf, len);
	if (result == (int)len)
		return 1;

	if (result < 0)
		thr_error_msg(rcpt_st->cwp, "write(%s): %s",
			      dcc_fnm2path(abs_nm, rcpt_st->user_log_nm),
			      ERROR_STR());
	else
		thr_error_msg(rcpt_st->cwp,
			      "write(%s)=%d instead of %d",
			      dcc_fnm2path(abs_nm, rcpt_st->user_log_nm),
			      result, (int)len);
	close(rcpt_st->user_log_fd);
	rcpt_st->user_log_fd = -1;
	return 0;
}



static void
user_log_print(RCPT_ST *rcpt_st, const char *p, ...)
{
	char logbuf[MAXHOSTNAMELEN*3];
	va_list args;

	va_start(args, p);
	vsnprintf(logbuf, sizeof(logbuf), p, args);
	va_end(args);
	user_log_write(rcpt_st, logbuf, 0);
}



static void
user_log_ck_write(void *rcpt_st0, const void *buf, u_int len)
{
	user_log_write((RCPT_ST *)rcpt_st0, buf, len);
}



static void
user_log_block(RCPT_ST *rcpt_st,	/* copy from main log file to this */
	       off_t start,		/* starting here */
	       off_t stop)		/* and ending here */
{
	DCC_PATH abs_nm;
		CMN_WORK *cwp;
	char buf[4096];
	off_t len;
	int result;

	cwp = rcpt_st->cwp;

	if (rcpt_st->user_log_fd < 0
	    || cwp->log_fd2 < 0)
		return;

	if (start == -1 || stop == -1) {
		thr_error_msg(cwp, "bogus user_log_block position");
		return;
	}

	if (-1 == lseek(cwp->log_fd2, start, SEEK_SET)) {
		thr_error_msg(cwp, "lseek(%s,%d,SEEK_SET): %s",
			      dcc_fnm2path(abs_nm, rcpt_st->user_log_nm),
			      (int)start, ERROR_STR());
		log_fd2_close(cwp, -2);
		return;
	}

	while ((len = stop - start) != 0) {
		if (len > (off_t)sizeof(buf))
			len = sizeof(buf);
		result = read(cwp->log_fd2, buf, len);
		if (result != len) {
			if (result < 0)
				thr_error_msg(cwp,
					      "user_log_block read(%s): %s",
					      dcc_fnm2path(abs_nm, cwp->log_nm),
					      ERROR_STR());
			else
				thr_error_msg(cwp, "user_log_block read(%s)=%d"
					      " instead of %d",
					      dcc_fnm2path(abs_nm, cwp->log_nm),
					      result, (int)len);
			log_fd2_close(cwp, -2);
			return;
		}
		if (!user_log_write(rcpt_st, buf, len))
			return;
		start += len;
	}
}



static void
user_log_msg(CMN_WORK *cwp, RCPT_ST *rcpt_st)
{
	DCC_PATH abs_nm;
	DCC_PATH rcpt_logdir;
	int log_mode;
	int i;

	cwp->ask_st |= ASK_ST_LOG_ONE;

	if (dcc_logdir[0] == '\0')
		return;

	snprintf(rcpt_logdir, sizeof(rcpt_logdir), "%s/log", rcpt_st->dir);
	i = dcc_logdir_ck(cwp->emsg, rcpt_logdir, &log_mode);
	if (i <= 0) {
		if (i < 0)
			thr_error_msg(cwp, "%s", cwp->emsg);
		return;
	}

	/* get an independent FD for the main log file that can be
	 * repositioned without affecting additional output to the main log. */
	if (cwp->log_fd2 < 0) {
		/* give up if things are already broken */
		if (cwp->log_fd2 != -1)
			return;
		/* Some systems don't synchronize the meta data among FDs for
		 * a file, causing the second FD to appear to be truncated. */
		if (fsync(cwp->log_fd) < 0)
			thr_error_msg(cwp, "fsync(%s): %s",
				      dcc_fnm2path(abs_nm, cwp->log_nm),
				      ERROR_STR());
		cwp->log_fd2 = open(cwp->log_nm, O_RDONLY, 0);
		if (cwp->log_fd2 < 0) {
			thr_error_msg(cwp, "open(%s): %s",
				      dcc_fnm2path(abs_nm, cwp->log_nm),
				      ERROR_STR());
			cwp->log_fd2 = -2;
			return;
		}
	}

	/* create the user's log file */
	rcpt_st->user_log_fd = dcc_mkstemp(cwp->emsg, rcpt_st->user_log_nm,
					   sizeof(rcpt_st->user_log_nm),
					   cwp->id, sizeof(cwp->id),
					   rcpt_logdir, DCC_FIN_LOG_PAT,
					   0, -1, log_mode & 0644);
	if (rcpt_st->user_log_fd < 0) {
		thr_error_msg(cwp, "%s", cwp->emsg);
		return;
	}

	user_log_block(rcpt_st,		/* copy envelope before env_To line */
		       0, cwp->log_pos_to_first);
	user_log_block(rcpt_st,		/* copy this env_To line */
		       rcpt_st->log_pos_to,
		       rcpt_st->fwd
		       ? rcpt_st->fwd->log_pos_to
		       : cwp->log_pos_to_end);
	user_log_block(rcpt_st,		/* copy the body of the message */
		       cwp->log_pos_to_end,
		       cwp->log_pos_white_first);
	user_log_block(rcpt_st,		/* copy whitelist error messages */
		       rcpt_st->log_pos_white,
		       rcpt_st->fwd
		       ? rcpt_st->fwd->log_pos_white
		       : cwp->log_pos_white_last);

	/* log external header, X-DCC header, and DCC results */
	if (!(cwp->ask_st & ASK_ST_LOCAL_WASSPAM)
	    && (cwp->ask_st & ASK_ST_SRVR_ISSPAM))
		USER_LOG_CAPTION(rcpt_st, DCC_XHDR_DCC_SPAM"\n");
	else if (cwp->ask_st & ASK_ST_SRVR_NOTSPAM)
		USER_LOG_CAPTION(rcpt_st, DCC_XHDR_DCC_OK"\n");
	if (cwp->ask_st & ASK_ST_REP_ISSPAM)
		USER_LOG_CAPTION(rcpt_st, DCC_XHDR_REP_SPAM"\n");
	USER_LOG_EOL(rcpt_st);
#ifdef USE_XFLTR
	if ((rcpt_st->flags & RCPT_ST_XFLTR_ON)
	    && cwp->xfltr_header[0] != '\0')
		user_log_print(rcpt_st, "%s\n", cwp->xfltr_header);
#endif
	if (cwp->header.buf[0] != '\0') {
		user_log_write(rcpt_st, cwp->header.buf, cwp->header.used);
		USER_LOG_EOL(rcpt_st);
	}

	/* log the checksums and their counts */
	dcc_print_cks(user_log_ck_write, rcpt_st,
		      cwp->local_tgts, &cwp->cks, rcpt_st->wtgts);
	USER_LOG_EOL(rcpt_st);
	dcc_print_grey(rcpt_st, rcpt_st->grey_result, rcpt_st->embargo_num,
		       0, rcpt_st->env_to,
		       rcpt_st->msg_sum, rcpt_st->triple_sum,
		       user_log_ck_write);
}



void
log_smtp_reply(CMN_WORK *cwp)
{
	LOG_CMN_CAPTION(cwp, DCC_XHDR_REJ_MSG);
	log_write(cwp, cwp->reply.rcode, 0);
	LOG_CMN_CAPTION(cwp, " ");
	log_write(cwp, cwp->reply.xcode, 0);
	LOG_CMN_CAPTION(cwp, " ");
	log_write(cwp, cwp->reply.buf, 0);
	LOG_CMN_EOL(cwp);
}



static void
user_log_smtp_reply(CMN_WORK *cwp, RCPT_ST *rcpt_st)
{
	USER_LOG_CAPTION(rcpt_st, DCC_XHDR_REJ_MSG);
	user_log_write(rcpt_st, cwp->reply.rcode, 0);
	USER_LOG_CAPTION(rcpt_st, " ");
	user_log_write(rcpt_st, cwp->reply.xcode, 0);
	USER_LOG_CAPTION(rcpt_st, " ");
	user_log_write(rcpt_st, cwp->reply.buf, 0);
	USER_LOG_EOL(rcpt_st);
}



/* process the message for each user to decide what to do with it */
void
users_process(CMN_WORK *cwp)
{
	RCPT_ST *rcpt_st;
	u_char need_eol;

	/* log what the DCC results and the headers in the common log file */
	if (!(cwp->ask_st & ASK_ST_LOCAL_WASSPAM)
	    && (cwp->ask_st & ASK_ST_SRVR_ISSPAM))
		LOG_CMN_CAPTION(cwp, DCC_XHDR_DCC_SPAM"\n");
	else if (cwp->ask_st & ASK_ST_SRVR_NOTSPAM)
		LOG_CMN_CAPTION(cwp, DCC_XHDR_DCC_OK"\n");
	if (cwp->ask_st & ASK_ST_REP_ISSPAM)
		LOG_CMN_CAPTION(cwp, DCC_XHDR_REP_SPAM"\n");
	LOG_CMN_EOL(cwp);
#ifdef USE_XFLTR
	if (cwp->xfltr_header[0] != '\0')
		thr_log_print(cwp, 0, "%s\n", cwp->xfltr_header);
#endif
	if (cwp->header.buf[0] != '\0') {
		log_write(cwp, cwp->header.buf, cwp->header.used);
		LOG_CMN_EOL(cwp);
	}

	/* log the checksums, DCC server counts and global whitelist values */
	dcc_print_cks(log_ck_write, cwp,
		      cwp->local_tgts, &cwp->cks, cwp->wtgts);
	LOG_CMN_EOL(cwp);

	/* mark recipients who won't receive it */
	need_eol = 0;
	for (rcpt_st = cwp->rcpt_st_first; rcpt_st; rcpt_st = rcpt_st->fwd) {
		/* decide whether it is spam for this target */
		if (!(rcpt_st->flags & RCPT_ST_WHITE)
		    && ((rcpt_st->flags & RCPT_ST_BLACK)
			|| ((cwp->ask_st & ASK_ST_SRVR_ISSPAM)
			    && !(rcpt_st->flags & RCPT_ST_DCC_OFF))
			|| ((cwp->ask_st & ASK_ST_DNSBL_ISSPAM)
			    && (rcpt_st->flags & RCPT_ST_DNSBL_ON))
			|| ((cwp->ask_st & ASK_ST_XFLTR_ISSPAM)
			    && (rcpt_st->flags & RCPT_ST_XFLTR_ON))))
			rcpt_st->flags |= RCPT_ST_ISSPAM;

		if (!(rcpt_st->flags & RCPT_ST_ISSPAM)) {
			++cwp->deliver_tgts;

		} else {
			++cwp->reject_tgts;
			/* Tell greylist to restore the embargo for targets
			 * that believe the message was spam and did not
			 * white- or blacklist it */
			if (rcpt_st->grey_result == ASK_GREY_EMBARGO
			    && rcpt_st->embargo_num == 0) {
				rcpt_st->grey_result = ASK_GREY_OFF;
			} else if (rcpt_st->grey_result == ASK_GREY_EMBARGO
				   || rcpt_st->grey_result==ASK_GREY_EMBARGO_END
				   || rcpt_st->grey_result == ASK_GREY_PASS) {
				rcpt_st->grey_result = ASK_GREY_SPAM;
				if (!dcc_grey_spam(cwp->emsg,
						   cwp->dcc_ctxt,
						   &cwp->cks,
						   rcpt_st->msg_sum,
						   rcpt_st->triple_sum))
					thr_error_msg(cwp, "%s", cwp->emsg);
			}
		}

		dcc_print_grey(cwp, rcpt_st->grey_result, rcpt_st->embargo_num,
			       &need_eol,
			       rcpt_st->env_to,
			       rcpt_st->msg_sum,
			       rcpt_st->triple_sum,
			       log_ck_write);
	}
	if (need_eol)
		LOG_CMN_EOL(cwp);


	/* do not embargo the message if no target wants it */
	if (cwp->deliver_tgts == 0)
		cwp->ask_st &= ~ASK_ST_GREY_EMBARGO;

	if (cwp->ask_st & ASK_ST_GREY_EMBARGO) {
		make_reply(cwp, &grey_reply);
		return;
	}

	/*  finished if we won't do anything about it */
	if (cwp->reject_tgts == 0
	    || cwp->action == CMN_IGNORE)
		return;

	/* make an SMTP rejection message unless we have already have one */
	if (cwp->reply.result)
		return;
	make_reply(cwp, &reject_reply);
}



/* after having checked each user or recipient,
 *	dispose of the message for each */
static void
user_log_result(CMN_WORK *cwp, RCPT_ST *rcpt_st, const char *result)
{
	/* create the per-user log file */
	if (rcpt_st->dir[0] != '\0'
	    && ((cwp->ask_st & ASK_ST_LOGIT)
		|| ((cwp->ask_st & ASK_ST_GREY_LOGIT)
		    && !(rcpt_st->flags & RCPT_ST_GREY_LOG_OFF))
		|| (rcpt_st->flags & RCPT_ST_LOG_ALL)))
		user_log_msg(cwp, rcpt_st);

	if (cwp->ask_st & ASK_ST_GREY_EMBARGO) {
		user_log_smtp_reply(cwp, rcpt_st);
		if (rcpt_st->embargo_num != 0) {
			user_log_print(rcpt_st,
				       DCC_XHDR_RESULT "%s #%d\n",
				       cwp->reply.result, rcpt_st->embargo_num);
		} else {
			user_log_print(rcpt_st, DCC_XHDR_RESULT"%s\n",
				       cwp->reply.result);
		}
		return;
	}

	if (!(rcpt_st->flags & RCPT_ST_ISSPAM)) {
		/* It is not spam for this target.
		 * If it was rejected late, e.g. by SMTP server for dccifd
		 * in proxy, mode, then log the rejection message */
		if (result) {
			if (cwp->reply.buf[0] != '\0') {
				USER_LOG_CAPTION(rcpt_st, DCC_XHDR_REJ_MSG);
				user_log_write(rcpt_st, cwp->reply.buf, 0);
			}
			USER_LOG_CAPTION(rcpt_st, DCC_XHDR_RESULT);
			user_log_write(rcpt_st, result, 0);
			USER_LOG_EOL(rcpt_st);
			return;
		}
		if (cwp->ask_st & (ASK_ST_LOCAL_WASSPAM
				   | ASK_ST_SRVR_ISSPAM
				   | ASK_ST_REP_ISSPAM)) {
			USER_LOG_CAPTION(rcpt_st, DCC_XHDR_RESULT);
			user_log_write(rcpt_st, DCC_XHDR_RESULT_I_A, 0);
			USER_LOG_EOL(rcpt_st);
			return;
		}
		if (rcpt_st->flags & RCPT_ST_GREY_END) {
			USER_LOG_CAPTION(rcpt_st,
					 DCC_XHDR_RESULT
					 DCC_XHDR_RESULT_ACCEPT
					 " after greylist embargo\n");
			return;
		}
		if (rcpt_st->flags & RCPT_ST_GREY_WHITE) {
			USER_LOG_CAPTION(rcpt_st,
					 DCC_XHDR_RESULT
					 DCC_XHDR_RESULT_ACCEPT
					 ";  greylist whitelist\n");
			return;
		}
		USER_LOG_CAPTION(rcpt_st,
				 DCC_XHDR_RESULT DCC_XHDR_RESULT_ACCEPT "\n");
		return;
	}

	/* It was spam for this target
	 * If some other target wanted the message, then we must discard it
	 * for this target.  Dccifd in SMTP proxy mode cannot discard for
	 * an individual target, and so must prevent the possibility */
	if (cwp->deliver_tgts != 0) {
		user_reject_discard(cwp, rcpt_st);
		if (cwp->action == CMN_DISCARD) {
			USER_LOG_CAPTION(rcpt_st, DCC_XHDR_RESULT"discard\n");
		} else {
			thr_log_print(cwp, 0,
				      DCC_XHDR_RESULT
				      DCC_XHDR_RESULT_DISCARD
				      " forced for %s\n",
				      rcpt_st->env_to);
			USER_LOG_CAPTION(rcpt_st,
					 DCC_XHDR_RESULT
					 DCC_XHDR_RESULT_DISCARD
					 " forced by other"
					 " target's whitelist\n");
		}
		return;
	}

	/* simple spam for all targets including this one */
	if (cwp->action == CMN_DISCARD) {
		USER_LOG_CAPTION(rcpt_st, DCC_XHDR_RESULT"discard\n");
		return;
	}

	user_log_smtp_reply(cwp, rcpt_st);
	user_log_print(rcpt_st, DCC_XHDR_RESULT"%s\n", cwp->reply.result);
}



/* log the consensus in each target's log file */
void
users_log_result(CMN_WORK *cwp,
		 const char *result)	/* 0 or reject by dccifd's server */
{
	RCPT_ST *rcpt_st;

	/* create individual log files and trim target list */
	for (rcpt_st = cwp->rcpt_st_first; rcpt_st; rcpt_st = rcpt_st->fwd) {
		user_log_result(cwp, rcpt_st, result);

		if (rcpt_st->user_log_fd >= 0) {
			if (0 > close(rcpt_st->user_log_fd))
				thr_error_msg(cwp, "close(user %s): %s",
					      rcpt_st->user_log_nm,
					      ERROR_STR());
			rcpt_st->user_log_fd = -1;
		}
	}

	log_fd2_close(cwp, -2);
}
